import numpy as np
import pandas as pd
from itertools import chain

from ..tools.tables import read_csv

def merge_grants(output):
    print('Merging all grant data')

    grant = read_csv(f'{output}/grant_grant.csv').set_index('patnum')
    firm = read_csv(f'{output}/grant_firm.csv').set_index('patnum')
    cite = read_csv(f'{output}/cite_stats.csv').set_index('patnum')
    assign = read_csv(f'{output}/assign_stats.csv').set_index('patnum')
    maint = read_csv(f'{output}/maint_maint.csv').set_index('patnum')

    grant = grant.join(firm)
    grant = grant.join(cite)
    grant = grant.join(assign)
    grant = grant.join(maint)

    fill_cols = ['n_cited', 'n_citing', 'n_self_cited', 'n_self_citing', 'n_trans', 'claims']
    grant[fill_cols] = grant[fill_cols].fillna(0).astype(np.int)

    int_cols = ['firm_num', 'last_maint']
    grant[int_cols] = grant[int_cols].astype('Int64')

    grant.drop('abstract', axis=1).to_csv(f'{output}/grant_info.csv')
    grant[['title', 'abstract']].to_csv(f'{output}/grant_text.csv')

def generate_firmyear(output, compustat=False):
    print('Generating all firm-years')

    total = []

    # patent applications
    apply = read_csv(f'{output}/apply_apply.csv', usecols=['appnum', 'appdate'])
    apply_firm = read_csv(f'{output}/apply_firm.csv').set_index('appnum')
    apply = apply.join(apply_firm, on='appnum', how='inner')
    apply['appyear'] = apply['appdate'].str.slice(0, 4).astype(np.int)

    apply_fy = apply.groupby(['firm_num', 'appyear']).size().rename('n_apply')
    apply_fy = apply_fy.rename_axis(index={'appyear': 'year'})
    total.append(apply_fy)

    # patent grants
    grant = read_csv(f'{output}/grant_info.csv', usecols=['patnum', 'pubdate', 'n_cited', 'n_citing', 'n_self_cited'])
    grant_firm = read_csv(f'{output}/grant_firm.csv').set_index('patnum')
    grant = grant.dropna(subset=['pubdate'], axis=0)
    grant['pubyear'] = grant['pubdate'].str.slice(0, 4).astype(np.int)
    grant = grant.join(grant_firm, on='patnum', how='inner')

    grant_groups = grant.groupby(['firm_num', 'pubyear'])
    grant_fy = grant_groups[['n_cited', 'n_citing', 'n_self_cited']].sum()
    grant_fy['n_grant'] = grant_groups.size()
    grant_fy = grant_fy.rename_axis(index={'pubyear': 'year'})
    total.append(grant_fy)

    # patent assignments
    assign = read_csv(f'{output}/assign_use.csv', usecols=['assignid', 'execdate'])
    assignor_firm = read_csv(f'{output}/assignor_firm.csv').set_index('assignid')
    assignee_firm = read_csv(f'{output}/assignee_firm.csv').set_index('assignid')
    assign = assign.join(assignor_firm.add_prefix('assignor_'), on='assignid', how='inner')
    assign = assign.join(assignee_firm.add_prefix('assignee_'), on='assignid', how='inner')
    assign['execyear'] = assign['execdate'].str.slice(0, 4).astype(np.int)

    assignor_fy = assign.groupby(['assignor_firm_num', 'execyear']).size().rename('n_source')
    assignor_fy = assignor_fy.rename_axis(index={'assignor_firm_num': 'firm_num', 'execyear': 'year'})
    total.append(assignor_fy)

    assignee_fy = assign.groupby(['assignee_firm_num', 'execyear']).size().rename('n_dest')
    assignee_fy = assignee_fy.rename_axis(index={'assignee_firm_num': 'firm_num', 'execyear': 'year'})
    total.append(assignee_fy)

    # compustat firms
    if compustat:
        compu = read_csv(f'{output}/compustat_compustat.csv')
        compu_firm = read_csv(f'{output}/compustat_firm.csv').set_index('compid')
        compu = compu.join(compu_firm, on='compid', how='inner')

        compu_grp = compu.groupby(['firm_num', 'year'])
        compu_fy = compu_grp[[
            'assets', 'capx', 'cash', 'cogs', 'deprec', 'income', 'employ',
            'intan', 'debt', 'revenue', 'sales', 'rnd', 'fcost', 'mktval'
        ]].sum()
        ind_info = compu_grp[['naics', 'sic']].first()
        compu_fy = compu_fy.join(ind_info)
        total.append(compu_fy)

    # comprehensive
    total = pd.concat(total, axis=1).reset_index()
    int_cols = ['n_apply', 'n_grant', 'n_cited', 'n_citing', 'n_self_cited', 'n_source', 'n_dest']
    total[int_cols] = total[int_cols].astype('Int64')

    total.to_csv(f'{output}/firmyear_info.csv', index=False, float_format='%.3f')

def firm_statistics(output):
    print('Finding firm statistics')

    # firm history statistics
    firmyear = read_csv(f'{output}/firmyear_info.csv', usecols=['firm_num', 'year', 'n_grant'])
    firm_groups = firmyear.groupby('firm_num')
    firm_life = pd.DataFrame({
        'year_min': firm_groups['year'].min(),
        'year_max': firm_groups['year'].max(),
        'tot_pats': firm_groups['n_grant'].sum()
    })
    firm_life['tot_pats'] = firm_life['tot_pats'].fillna(0).astype(np.int)
    firm_life['life_span'] = firm_life['year_max'] - firm_life['year_min'] + 1

    # load in ipc info
    grant = read_csv(f'{output}/grant_info.csv', usecols=['firm_num', 'ipc'])
    grant = grant.dropna(subset=['firm_num'])
    grant['firm_num'] = grant['firm_num'].astype('Int64')
    grant['ipc4'] = grant['ipc'].str.slice(0, 4)

    # get modal ipc4 info
    count_ipc = grant.groupby(['firm_num', 'ipc4']).size().rename('count_ipc4')
    firm_ipc = count_ipc.reset_index(level='firm_num').groupby('firm_num')['count_ipc4']
    mode_ipc = firm_ipc.idxmax().rename('mode_ipc4')
    mode_ipc_count = firm_ipc.max().rename('mode_ipc4_count').astype('Int64')
    firm_life = firm_life.join(mode_ipc)
    firm_life = firm_life.join(mode_ipc_count)
    firm_life['mode_ipc4_frac'] = firm_life['mode_ipc4_count']/firm_life['tot_pats']
    firm_life = firm_life.drop('mode_ipc4_count', axis=1)

    firm_life.to_csv(f'{output}/firm_life.csv', float_format='%.3f')

def patent_stocks(output):
    print('Constructing patent stocks')

    # load firm data
    firmyear_info = read_csv(f'{output}/firmyear_info.csv')
    firm_info = read_csv(f'{output}/firm_life.csv', usecols=['firm_num', 'year_min', 'year_max', 'life_span'])

    # make (firm_num, year) index
    fnum_set = firm_info['firm_num']
    year_min = firm_info['year_min']
    year_max = firm_info['year_max']
    life_span = firm_info['life_span']
    all_fnums = np.array(list(chain(*[[fnum]*life for fnum, life in zip(fnum_set, life_span)])), dtype=np.int)
    all_years = np.array(list(chain(*[range(x, y+1) for x, y in zip(year_min, year_max)])), dtype=np.int)
    fy_all = pd.DataFrame({
        'firm_num': all_fnums,
        'year': all_years
    })

    datf_idx = fy_all.merge(firmyear_info, how='left', on=['firm_num', 'year'])
    int_cols = ['n_apply', 'n_grant', 'n_citing', 'n_cited', 'n_self_cited', 'n_source', 'n_dest']
    datf_idx = datf_idx.fillna({c: 0 for c in int_cols})
    datf_idx[int_cols] = datf_idx[int_cols].astype(np.int)

    # merge in overall firm info
    datf_idx = datf_idx.merge(firm_info[['firm_num', 'year_min']], how='left', on='firm_num')
    datf_idx['age'] = datf_idx['year'] - datf_idx['year_min']
    datf_idx = datf_idx.drop('year_min', axis=1)

    # aggregate stocks
    firm_groups = datf_idx.groupby('firm_num')
    datf_idx['stock'] = firm_groups['n_apply'].cumsum()

    # write new frame to disk
    datf_idx.to_csv(f'{output}/firmyear_index.csv', index=False, float_format='%.3f')

def merge_firms(output, compustat=False):
    merge_grants(output)
    generate_firmyear(output, compustat=compustat)
    firm_statistics(output)
    patent_stocks(output)



#!/usr/bin/env python3

import os
import sys
import fire
from importlib import resources
from pathlib import Path

from . import (
    fetch_many, concat_tables, parse_apply, parse_grant, parse_assign,
    parse_assign, parse_maint, parse_tmapply, parse_compustat, cluster_firms,
    prune_assign, aggregate_cites, merge_firms
)

# parser dispatcher
parsers = {
    'apply': parse_apply,
    'grant': parse_grant,
    'assign': parse_assign,
    'maint': parse_maint,
    'tmapply': parse_tmapply,
    'compustat': parse_compustat,
}

# get path with fallbacks
def get_path(path, env):
    if path is None:
        if env in os.environ:
            return Path(os.environ[env])
        else:
            return
    else:
        return Path(path)

# read list files
def get_lines(fpath):
    return [
        s.strip() for s in fpath.read_text().split('\n') if len(s) > 0
    ]

class Main:
    def __init__(self, datadir=None, metadir=None):
        self.datapath = get_path(datadir, 'FASTPAT_DATADIR')
        self.metapath = get_path(metadir, 'FASTPAT_METADIR')

        if self.datapath is None:
            print('Error: must specify "datadir" path')
            sys.exit()

        if self.metapath is None:
            self.metapath = resources.files('fastpat') / 'meta'

    def stat(self):
        print(f'datapath: {self.datapath}')
        print(f'metapath: {self.metapath}')

    def fetch(self, ftype, files=None, delay=10, overwrite=False, unzip=False, dryrun=False):
        if files is None:
            fpath = self.metapath / f'{ftype}_files.txt'
        else:
            fpath = Path(files)
        flist = get_lines(fpath)

        output = self.datapath / 'raw' / ftype
        fetch_many(flist, output, delay=delay, overwrite=overwrite, unzip=unzip, dryrun=dryrun)

    def parse(self, ftype, path=None, concat=True, overwrite=False, dryrun=False, threads=10):
        if path is None:
            path = self.datapath / 'raw' / ftype
        pardir = self.datapath / 'parsed' / ftype
        tabdir = self.datapath / 'tables'

        if ftype in parsers:
            parsers[ftype](path, pardir, overwrite=overwrite, dryrun=dryrun, threads=threads)
            if concat:
                concat_tables(pardir, tabdir, ftype)
        else:
            print(f'Error: unknown data source "{ftype}"')

    def firms(self, action, sources=None, compustat=False):
        tabdir = self.datapath / 'tables'

        if action == 'assign':
            prune_assign(tabdir)
        elif action == 'cluster':
            cluster_firms(tabdir, sources=sources)
        elif action == 'cites':
            aggregate_cites(tabdir)
        elif action == 'merge':
            merge_firms(tabdir, compustat=compustat)
        else:
            print(f'Error: unknown firm action "{action}"')

def main():
    fire.Fire(Main)



import re
import os
import glob
from collections import defaultdict
from traceback import print_exc
from multiprocessing import Pool

from ..tools.parse import *
from ..tools.tables import ChunkWriter, DummyWriter

def parse_apply_gen2(elem, fname):
    pat = defaultdict(str)
    pat['gen'] = 2
    pat['file'] = fname

    # top-level section
    bib = elem.find('subdoc-bibliographic-information')

    # publication data
    pub = bib.find('document-id')
    if pub is not None:
        pat['pubnum'] = get_text(pub, 'doc-number')
        pat['pubdate'] = get_text(pub, 'document-date')

    # application data
    app = bib.find('domestic-filing-data')
    if app is not None:
        pat['appnum'] = get_text(app, 'application-number/doc-number')
        pat['appdate'] = get_text(app, 'filing-date')

    # title
    tech = bib.find('technical-information')
    pat['title'] = get_text(tech, 'title-of-invention')

    # ipc code
    pat['ipcs'] = []
    ipcsec = tech.find('classification-ipc')
    if ipcsec is not None:
        pat['ipcver'] = get_text(ipcsec, 'classification-ipc-edition').lstrip('0')
        pat['ipcs'] = [ip for ip in gen2_ipc(ipcsec)]

    # assignee name
    pat['appname'] = get_text(bib, 'assignee/organization-name')

    # first inventor address
    resid = bib.find('inventors/first-named-inventor/residence')
    if resid is not None:
        address = resid.find('residence-us')
        if address is None:
            address = resid.find('residence-non-us')
        if address is not None:
            pat['city'] = get_text(address, 'city')
            pat['state'] = get_text(address, 'state')
            pat['country'] = get_text(address, 'country-code')

    # abstract
    abst = elem.find('subdoc-abstract')
    if abst is not None:
        pat['abstract'] = raw_text(abst, sep=' ')

    # roll it in
    return pat

def parse_apply_gen3(elem, fname):
    pat = defaultdict(str)
    pat['gen'] = 3
    pat['file'] = fname

    # top-level section
    bib = elem.find('us-bibliographic-data-application')
    pubref = bib.find('publication-reference')
    appref = bib.find('application-reference')

    # published patent
    pubinfo = pubref.find('document-id')
    pat['pubnum'] = get_text(pubinfo, 'doc-number')
    pat['pubdate'] = get_text(pubinfo, 'date')

    # filing date
    pat['appnum'] = get_text(appref, 'document-id/doc-number')
    pat['appdate'] = get_text(appref, 'document-id/date')
    pat['appname'] = get_text(bib, 'assignees/assignee/addressbook/orgname')

    # title
    pat['title'] = get_text(bib, 'invention-title')

    # ipc code
    pat['ipcs'] = []
    ipcsec = bib.find('classification-ipc')
    if ipcsec is not None:
        pat['ipcver'] = get_text(ipcsec, 'edition').lstrip('0')
        pat['ipcs'] = [ip for ip in gen3a_ipc(ipcsec)]
    else:
        ipcsec = bib.find('classifications-ipcr')
        if ipcsec is not None:
            pat['ipcver'] = get_text(ipcsec, 'classification-ipcr/ipc-version-indicator/date')
            pat['ipcs'] = [ip for ip in gen3r_ipc(ipcsec)]

    # first inventor address
    address = bib.find('parties/applicants/applicant/addressbook/address')
    if address is None:
        address = bib.find('us-parties/us-applicants/us-applicant/addressbook/address')
    if address is not None:
        pat['city'] = get_text(address, 'city')
        pat['state'] = get_text(address, 'state')
        pat['country'] = get_text(address, 'country')

    # abstract
    abspar = elem.find('abstract')
    if abspar is not None:
        pat['abstract'] = raw_text(abspar, sep=' ')

    # roll it in
    return pat

# table schema
schema_apply = {
    'appnum': 'str', # Patent number
    'appdate': 'str', # Application date
    'appname': 'str', # Assignee name
    'pubnum': 'str', # Publication number
    'pubdate': 'str', # Publication date
    'ipc': 'str', # Main IPC code
    'ipcver': 'str', # IPC version info
    'city': 'str', # Assignee city
    'state': 'str', # State code
    'country': 'str', # Assignee country
    'title': 'str', # Title
    'abstract': 'str', # Abstract
    'gen': 'int', # USPTO data format
    'file': 'str', # path to source file
}

schema_ipc = {
    'patnum': 'str', # Patent number
    'ipc': 'str', # IPC code
    'rank': 'int', # Order listed
    'version': 'str' # IPC version
}

# chunking express
def store_patent(pat, chunker_pat, chunker_ipc):
    an, iv = pat['appnum'], pat['ipcver']

    # store ipcs
    for j, ipc in enumerate(pat['ipcs']):
        if j == 0: pat['ipc'] = ipc
        chunker_ipc.insert(an, ipc, j, iv)

    # store patent
    chunker_pat.insert(*(pat.get(k, '') for k in schema_apply))

# file level
def parse_file(fpath, output, display=0, overwrite=False, dryrun=False):
    fdir, fname = os.path.split(fpath)
    ftag, fext = os.path.splitext(fname)

    opath_apply = os.path.join(output, f'apply_{ftag}.csv')
    opath_ipc = os.path.join(output, f'ipc_{ftag}.csv')

    complete = os.path.exists(opath_apply) and os.path.exists(opath_ipc)
    if not overwrite and complete:
        print(f'{ftag}: Skipping')
        return

    if fname.startswith('pab'):
        parser = lambda fp: parse_wrapper(fp, 'patent-application-publication', parse_apply_gen2)
    elif fname.startswith('ipab'):
        parser = lambda fp: parse_wrapper(fp, 'us-patent-application', parse_apply_gen3)
    else:
        raise Exception(f'{ftag}: Unknown format')

    if dryrun:
        chunker_apply = DummyWriter()
        chunker_ipc = DummyWriter()
    else:
        chunker_apply = ChunkWriter(opath_apply, schema=schema_apply)
        chunker_ipc = ChunkWriter(opath_ipc, schema=schema_ipc)

    # parse it up
    try:
        print(f'{ftag}: Starting')

        i = 0
        for pat in parser(fpath):
            i += 1

            store_patent(pat, chunker_apply, chunker_ipc)

            # output
            if display > 0 and i % display == 0:
                spat = {k: pat.get(k, '') for k in schema_apply}
                print(
                    'an = {appnum:10.10s}, fd = {appdate:10.10s}, ti = {title:30.30s}, '
                    'on = {appname:30.30s}, ci = {city:15.15s}, st = {state:2s}, '
                    'ct = {country:2s}'.format(**spat)
                )

        # commit to db and close
        chunker_apply.commit()
        chunker_ipc.commit()

        print(f'{ftag}: Parsed {i} patents')
    except Exception as e:
        print(f'{ftag}: EXCEPTION OCCURRED!')
        print_exc()

        chunker_apply.delete()
        chunker_ipc.delete()

# main entry point
def parse_many(files, output, threads=10, display=1_000, overwrite=False, dryrun=False):
    # needed for multiprocess
    global parse_file_opts

    # collect files
    if type(files) is str or isinstance(files, os.PathLike):
        file_list = (
            sorted(glob.glob(f'{files}/pab*.xml')) +
            sorted(glob.glob(f'{files}/ipab*.xml'))
        )
    else:
        file_list = files

    # ensure output dir
    if not dryrun and not os.path.exists(output):
        print(f'Creating directory {output}')
        os.makedirs(output)

    # apply options
    def parse_file_opts(fpath):
        parse_file(fpath, output, display=display, overwrite=overwrite, dryrun=dryrun)

    # parse files
    with Pool(threads) as pool:
        pool.map(parse_file_opts, file_list, chunksize=1)



##
## common parsing tools
##

import os
import re
import numpy as np
import pandas as pd
from lxml.etree import XMLPullParser

##
## xml parsing
##

# get descendent text
def get_text(parent, tag, default=''):
    child = parent.find(tag)
    if child is None:
        return default
    elif child.text is None:
        return default
    else:
        return child.text.strip().lower()

# get all text of node
def raw_text(par, sep=''):
    return sep.join(par.itertext()).strip().lower()

# preserve memory
def clear(elem):
    elem.clear()
    while elem.getprevious() is not None:
        del elem.getparent()[0]

# parse mangled xml
def parse_wrapper(fpath, main_tag, parser):
    _, fname = os.path.split(fpath)
    pp = XMLPullParser(tag=main_tag, events=['end'], recover=True)
    def parse_all():
        for _, pat in pp.read_events():
            yield parser(pat, fname)
            clear(pat)

    with open(fpath, errors='ignore') as f:
        pp.feed('<root>\n')
        for line in f:
            if line.startswith('<?xml'):
                yield from parse_all()
            elif line.startswith('<!DOCTYPE') or line.startswith('<!ENTITY') or line.startswith(']>'):
                pass
            else:
                pp.feed(line)
        else:
            pp.feed('</root>\n')
            yield from parse_all()

##
## patnum parsers
##

# standard way to pruce patent numbers (allows for all types)
def prune_patnum(pn, maxlen=7):
    ret = re.match(r'([a-zA-Z]{1,2}|0)?([0-9]+)', pn)
    if ret is None:
        prefix = ''
        patnum = pn
    else:
        prefix, patnum = ret.groups()
        prefix = '' if (prefix is None or prefix == '0') else prefix
    patnum = patnum[:maxlen].lstrip('0')
    return prefix + patnum

##
## ipc parsers
##

# early grant only (1, 1.5)
def pad_ipc(ipc):
    if len(ipc) >= 8:
        return ipc[:4] + ipc[4:7].replace(' ', '0') + '/' + ipc[7:]
    else:
        return ipc

# grant only (1.5)
def gen15_ipc(ipcsec):
    yield get_text(ipcsec, 'B511/PDAT')
    for ipc in ipcsec.findall('B512'):
        yield get_text(ipc, 'PDAT')

# apply only (2)
def gen2_ipc(ipcsec):
    yield get_text(ipcsec, 'classification-ipc-primary/ipc')
    for ipc in ipcsec.findall('classification-ipc-secondary'):
        yield get_text(ipc, 'ipc')

# apply and grant (3)
def gen3a_ipc(ipcsec):
    yield get_text(ipcsec, 'main-classification')
    for ipc in ipcsec.findall('further-classification'):
        yield ipc.text or ''

def gen3g_ipc(ipcsec):
    yield get_text(ipcsec, 'main-classification')
    for ipc in ipcsec.findall('further-classification'):
        yield ipc.text or ''

# apply and grant (3)
def gen3r_ipc(ipcsec):
    for ipc in ipcsec.findall('classification-ipcr'):
        yield get_text(ipc, 'section') + get_text(ipc, 'class') + get_text(ipc, 'subclass') \
            + get_text(ipc, 'main-group').zfill(3) + '/' + get_text(ipc, 'subgroup')

##
## cite parsers
##

# grant (2)
def gen2_cite(refs):
    for cite in refs.findall('B561'):
        yield get_text(cite, 'PCIT/DOC/DNUM/PDAT')

# grant (3)
def gen3_cite(refs, prefix):
    for cite in refs.findall(prefix+'citation/patcit/document-id'):
        natl = get_text(cite, 'country')
        kind = get_text(cite, 'kind')
        pnum = get_text(cite, 'doc-number')
        if natl == 'us' and kind != '00': # US granted patents only
            yield pnum

##
## assign parsers
##

def gen3_assign(patents):
    for doc in patents.findall('patent-property/document-id'):
        kind = get_text(doc, 'kind')
        pnum = get_text(doc, 'doc-number')
        if kind.startswith('b'):
            yield pnum

# detect organization type
ORG_CORP = 0
ORG_NONP = 1
ORG_INDV = 2

LEN_CUT = 30

corp_keys = ['corp', 'co', 'inc', 'llc', 'lp', 'plc', 'ltd', 'limited', 'company', 'corporation', 'incorporated', 'international', 'systems', 'sa', 'oy', 'consulting', 'bank', 'gmbh', 'kabushiki', 'kaisha', 'bv', 'nv', 'sl', 'aktiengesellschaft', 'maschinenfabrik', 'ab', 'ag', 'as', 'spa', 'hf', 'societe', 'associates', 'business', 'industries', 'group', 'kk', 'laboratories', 'works', 'studio', 'telecom', 'investments', 'consultants', 'electronics', 'technologies', 'microsystems', 'multimedia', 'networks', 'technology', 'partnership', 'electric', 'components', 'automotive', 'instruments', 'communication', 'enterprises', 'network', 'engineering', 'designs', 'sciences', 'partners', 'aktiengellschaft', 'venture', 'aerospace', 'pharmaceuticals', 'design', 'medical', 'products', 'pharma', 'energy', 'solutions', 'france', 'isreal', 'product', 'plastics', 'communications', 'kgaa', 'sas', 'cellular', 'gesellschaft', 'se', 'holdings', 'kg', 'srl', 'chimie']
nonp_keys = ['institute', 'university', 'hospital', 'foundation', 'college', 'research', 'administration', 'recherche', 'department', 'trust', 'association', 'ministry', 'laboratory', 'board', 'office', 'univ', 'ecole', 'secretary', 'universidad', 'society', 'universiteit', 'centre', 'center', 'national', 'school', 'institut', 'institutes', 'universite']

punc_re = re.compile(r'[0-9&()]')
spac_re = re.compile(r'[ ,]')
corp_re = re.compile('\\b('+'|'.join(corp_keys)+')\\b')
nonp_re = re.compile('\\b('+'|'.join(nonp_keys)+')\\b')

def org_type(name):
    name = name.replace('.', '')
    name = name.replace('/', '')
    has_corp = corp_re.search(name) != None
    has_nonp = nonp_re.search(name) != None
    has_punc = punc_re.search(name) != None
    has_spac = spac_re.search(name) != None
    long_name = len(name) > LEN_CUT
    if has_corp or has_punc or not has_spac or long_name:
        return ORG_CORP
    elif has_nonp:
        return ORG_NONP
    else:
        return ORG_INDV

# detect conveyance type
CONV_ASSIGN = 0
CONV_LICENSE = 1
CONV_MERGER = 2
CONV_OTHER = 3

# detect if a conveyance is not a name/address change or security agreement
other_keys = ['change', 'secur', 'correct', 'release', 'lien', 'update', 'nunc', 'collat']
other_re = re.compile('|'.join(other_keys))

def convey_type(convey):
    if other_re.search(convey) != None:
        return CONV_OTHER
    elif convey.find('assign') != -1:
        return CONV_ASSIGN
    elif convey.find('license') != -1:
        return CONV_LICENSE
    elif convey.find('merge') != -1:
        return CONV_MERGER
    else:
        return CONV_OTHER



import re
import os
import glob
from collections import defaultdict
from traceback import print_exc
from multiprocessing import Pool
from lxml.etree import iterparse

from ..tools.parse import *
from ..tools.tables import ChunkWriter, DummyWriter

def parse_tmapply(elem, fname):
    tma = defaultdict(str)
    tma['file'] = fname

    # top-level section
    tma['serial'] = get_text(elem, 'serial-number')

    # header information
    head = elem.find('case-file-header')
    if head is not None:
        tma['regdate'] = get_text(head, 'registration-date')
        tma['filedate'] = get_text(head, 'filing-date')

    # case file statements
    stats = []
    for st in elem.findall('case-file-statements/case-file-statement'):
        tc = get_text(st, 'type-code')
        if tc.startswith('gs'):
            tx = get_text(st, 'text')
            stats.append((tc, tx))

    # summarize
    tma['n_stat'] = len(stats)
    tma['gs_codes'] = '; '.join([tc for tc, _ in stats])
    tma['statement'] = '; '.join([tx for _, tx in stats])

    # classifications
    uclas = []
    iclas = []
    for cl in elem.findall('classifications/classification'):
        uc = [u.text for u in cl.findall('us-code')]
        ic = [i.text for i in cl.findall('international-code')]
        uclas += uc
        iclas += ic
    tma['us_class'] = '; '.join(uclas)
    tma['int_class'] = '; '.join(iclas)

    # owners
    owns = []
    for ow in elem.findall('case-file-owners/case-file-owner'):
        nm = get_text(ow, 'party-name')
        owns.append(nm)
    tma['owners'] = '; '.join(dict.fromkeys(owns))

    # roll it in
    return tma

def store_tmapply(tma, chunker):
    chunker.insert(*(tma.get(k, '') for k in schema_tmapply))

# table schema
schema_tmapply = {
    'serial': 'str', # trademark number
    'regdate': 'str', # registration date
    'filedate': 'str', # filing date
    'n_stat': 'int', # number of GS statements
    'gs_codes': 'str', # all GS codes in statements
    'statement': 'str', # concatenated GS statements
    'us_class': 'str', # US classifications
    'int_class': 'str', # international classifications
    'owners': 'str', # all owner names
    'file': 'str', # path to source file
}

# file level
def parse_file(fpath, output, overwrite=False, dryrun=False, display=0):
    fdir, fname = os.path.split(fpath)
    ftag, fext = os.path.splitext(fname)

    opath_tmapply = os.path.join(output, f'tmapply_{ftag}.csv')

    if not overwrite and os.path.exists(opath_tmapply):
        print(f'{ftag}: Skipping')
        return

    if dryrun:
        chunker_tma = DummyWriter()
    else:
        chunker_tma = ChunkWriter(opath_tmapply, schema=schema_tmapply)

    # parse it up
    try:
        print(f'{ftag}: Starting')

        i = 0
        for event, elem in iterparse(fpath, tag='case-file', events=['end'], recover=True):
            i += 1

            # parse and store
            tma = parse_tmapply(elem, fname)
            store_tmapply(tma, chunker_tma)
            clear(elem)

            # output
            if display > 0 and i % display == 0:
                stma = {k: tma.get(k, '') for k in schema_tmapply}
                print(
                    'fn = {file:20.20s}, sn = {serial:10.10s}, rd = {regdate:10.10s}, '
                    'ic = {int_class:10.10s}, gs = {gs_codes:15.15s}, st = {statement:50.50s}, '
                    'ow = {owners:30.30s}'.format(**stma)
                )

        # commit to db and close
        chunker_tma.commit()

        print(f'{ftag}: Parsed {i} trademarks')
    except Exception as e:
        print(f'{ftag}: EXCEPTION OCCURRED!')
        print_exc()

        chunker_tma.delete()

# main entry point
def parse_many(files, output, threads=10, display=1_000, overwrite=False, dryrun=False):
    # needed for multiprocess
    global parse_file_opts

    # collect files
    if type(files) is str or isinstance(files, os.PathLike):
        file_list = sorted(glob.glob(f'{files}/apc*.xml'))
    else:
        file_list = files

    # ensure output dir
    if not dryrun and not os.path.exists(output):
        print(f'Creating directory {output}')
        os.makedirs(output)

    # apply options
    def parse_file_opts(fpath):
        parse_file(fpath, output, display=display, overwrite=overwrite, dryrun=dryrun)

    # parse files
    with Pool(threads) as pool:
        pool.map(parse_file_opts, file_list, chunksize=1)



import re
import os
import glob
from collections import defaultdict
from traceback import print_exc
from itertools import chain
from multiprocessing import Pool

from ..tools.parse import *
from ..tools.tables import ChunkWriter, DummyWriter

# parse it up
def parse_grant_gen1(fname):
    pat = None
    sec = None
    tag = None
    ipcver = None
    for nline in chain(open(fname, encoding='latin1', errors='ignore'), ['PATN']):
        # peek at next line
        ntag, nbuf = nline[:4].rstrip(), nline[5:-1].rstrip().lower()
        if tag is None:
            tag = ntag
            buf = nbuf
            continue
        if ntag == '':
            buf += nbuf
            continue

        # regular tags
        if tag == 'PATN':
            if pat is not None:
                pat['appnum'] = src + apn
                yield pat
            pat = defaultdict(str)
            sec = 'PATN'
            pat['gen'] = 1
            _, pat['file'] = os.path.split(fname)
            pat['ipcs'] = []
            pat['cites'] = []
            src, apn = '', ''
        elif tag in ['INVT', 'ASSG', 'PRIR', 'CLAS', 'UREF', 'FREF', 'OREF', 'LREP', 'PCTA', 'ABST']:
            sec = tag
        elif tag in ['PAL', 'PAR', 'PAC', 'PA0', 'PA1']:
            if sec == 'ABST':
                if 'abstract' not in pat:
                    pat['abstract'] = buf
                else:
                    pat['abstract'] += '\n' + buf
        elif tag == 'WKU':
            if sec == 'PATN':
                pat['patnum'] = prune_patnum(buf)
        elif tag == 'SRC':
            if sec == 'PATN':
                src = '29' if buf == 'd' else buf.zfill(2) # design patents get series code 29
        elif tag == 'APN':
            if sec == 'PATN':
                apn = buf[:6]
        elif tag == 'ISD':
            if sec == 'PATN':
                pat['pubdate'] = buf
        elif tag == 'APD':
            if sec == 'PATN':
                pat['appdate'] = buf
        elif tag == 'ICL':
            if sec == 'CLAS':
                pat['ipcs'].append(pad_ipc(buf))
        elif tag == 'EDF':
            if sec == 'CLAS':
                pat['ipcver'] = buf
        elif tag == 'TTL':
            if sec == 'PATN':
                pat['title'] = buf
        elif tag == 'NCL':
            if sec == 'PATN':
                pat['claims'] = buf
        elif tag == 'NAM':
            if sec == 'ASSG':
                pat['owner'] = buf
        elif tag == 'CTY':
            if sec == 'ASSG':
                pat['city'] = buf
        elif tag == 'STA':
            if sec == 'ASSG':
                pat['state'] = buf
                pat['country'] = 'us'
        elif tag == 'CNT':
            if sec == 'ASSG':
                pat['country'] = buf[:2]
        elif tag == 'PNO':
            if sec == 'UREF':
                pat['cites'].append(prune_patnum(buf))

        # stage next tag and buf
        tag = ntag
        buf = nbuf

def parse_grant_gen2(elem, fname):
    pat = defaultdict(str)
    pat['gen'] = 2
    pat['file'] = fname

    # top-level section
    bib = elem.find('SDOBI')

    # published patent
    pubref = bib.find('B100')
    pat['patnum'] = prune_patnum(get_text(pubref, 'B110/DNUM/PDAT'))
    pat['pubdate'] = get_text(pubref, 'B140/DATE/PDAT')

    # filing date
    appref = bib.find('B200')
    pat['appnum'] = get_text(appref, 'B210/DNUM/PDAT')
    pat['appdate'] = get_text(appref, 'B220/DATE/PDAT')

    # reference info
    patref = bib.find('B500')
    ipcsec = patref.find('B510')
    if ipcsec is not None:
        pat['ipcver'] = get_text(ipcsec, 'B516/PDAT')
        pat['ipcs'] = [pad_ipc(ip) for ip in gen15_ipc(ipcsec)]
    else:
        pat['ipcs'] = []
    pat['title'] = get_text(patref, 'B540/STEXT/PDAT')
    pat['claims'] = get_text(patref, 'B570/B577/PDAT')

    # citations
    refs = patref.find('B560')
    if refs is not None:
        pat['cites'] = [prune_patnum(pn) for pn in gen2_cite(refs)]
    else:
        pat['cites'] = []

    # applicant name and address
    ownref = bib.find('B700/B730/B731/PARTY-US')
    if ownref is not None:
        pat['owner'] = get_text(ownref, 'NAM/ONM/STEXT/PDAT')
        address = ownref.find('ADR')
        if address is not None:
            pat['city'] = get_text(address, 'CITY/PDAT')
            pat['state'] = get_text(address, 'STATE/PDAT')
            pat['country'] = get_text(address, 'CTRY/PDAT')

    # abstract
    abspars = elem.findall('SDOAB/BTEXT/PARA')
    if len(abspars) > 0:
        pat['abstract'] = '\n'.join([raw_text(e) for e in abspars])

    # roll it in
    return pat

def parse_grant_gen3(elem, fname):
    pat = defaultdict(str)
    pat['gen'] = 3
    pat['file'] = fname

    # top-level section
    bib = elem.find('us-bibliographic-data-grant')
    pubref = bib.find('publication-reference')
    appref = bib.find('application-reference')

    # published patent
    pubinfo = pubref.find('document-id')
    pat['patnum'] = prune_patnum(get_text(pubinfo, 'doc-number'), maxlen=8)
    pat['pubdate'] = get_text(pubinfo, 'date')

    # filing date
    appinfo = appref.find('document-id')
    pat['appnum'] = get_text(appinfo, 'doc-number')
    pat['appdate'] = get_text(appinfo, 'date')

    # title
    pat['title'] = get_text(bib, 'invention-title')

    # ipc code
    pat['ipcs'] = []
    ipcsec = bib.find('classification-ipc')
    if ipcsec is not None:
        pat['ipcver'] = get_text(ipcsec, 'edition')
        pat['ipcs'] = [pad_ipc(ip) for ip in gen3g_ipc(ipcsec)]
    else:
        ipcsec = bib.find('classifications-ipcr')
        if ipcsec is not None:
            pat['ipcver'] = get_text(ipcsec, 'classification-ipcr/ipc-version-indicator/date')
            pat['ipcs'] = [ip for ip in gen3r_ipc(ipcsec)]

    # claims
    pat['claims'] = get_text(bib, 'number-of-claims')

    # citations
    refs = bib.find('references-cited')
    prefix = ''
    if refs is None:
        refs = bib.find('us-references-cited')
        prefix = 'us-'
    if refs is not None:
        pat['cites'] = [prune_patnum(pn, maxlen=8) for pn in gen3_cite(refs, prefix)]
    else:
        pat['cites'] = []

    # applicant name and address
    assignee = bib.find('assignees/assignee/addressbook')
    if assignee is not None:
        pat['owner'] = get_text(assignee, 'orgname')
        address = assignee.find('address')
        if address is not None:
            pat['city'] = get_text(address, 'city')
            pat['state'] = get_text(address, 'state')
            pat['country'] = get_text(address, 'country')

    # abstract
    abspar = elem.find('abstract')
    if abspar is not None:
        pat['abstract'] = raw_text(abspar, sep=' ')

    # roll it in
    return pat

# table schemas
schema_grant = {
    'patnum': 'str', # Patent number
    'pubdate': 'str', # Publication date
    'appnum': 'str', # Application number
    'appdate': 'str', # Publication date
    'ipc': 'str', # Main IPC code
    'ipcver': 'str', # IPC version info
    'city': 'str', # Assignee city
    'state': 'str', # State code
    'country': 'str', # Assignee country
    'owner': 'str', # Assignee name
    'claims': 'int', # Independent claim
    'title': 'str', # Title
    'abstract': 'str', # Abstract
    'gen': 'int', # USPTO data format
    'file': 'str', # path to source file
}

schema_ipc = {
    'patnum': 'str', # Patent number
    'ipc': 'str', # IPC code
    'rank': 'int', # Order listed
    'version': 'str' # IPC version
}

schema_cite = {
    'src': 'str', # Source patent (citer)
    'dst': 'str' # Destination patent (citee)
}

# patent adder
def store_patent(pat, chunker_grant, chunker_ipc, chunker_cite):
    pn, iv = pat['patnum'], pat['ipcver']

    # store cites
    for cite in pat['cites']:
        chunker_cite.insert(pn, cite)

    # store ipcs
    for j, ipc in enumerate(pat['ipcs']):
        if j == 0:
            pat['ipc'] = ipc
        chunker_ipc.insert(pn, ipc, j, iv)

    # store patent
    chunker_grant.insert(*(pat.get(k, '') for k in schema_grant))

# file level
def parse_file(fpath, output, display=0, overwrite=False, dryrun=False):
    fdir, fname = os.path.split(fpath)
    ftag, fext = os.path.splitext(fname)

    opath_grant = os.path.join(output, f'grant_{ftag}.csv')
    opath_ipc = os.path.join(output, f'ipc_{ftag}.csv')
    opath_cite = os.path.join(output, f'cite_{ftag}.csv')

    complete = (
        os.path.exists(opath_grant) and
        os.path.exists(opath_ipc) and
        os.path.exists(opath_cite)
    )
    if not overwrite and complete:
        print(f'{ftag}: Skipping')
        return

    if dryrun:
        chunker_grant = DummyWriter()
        chunker_ipc = DummyWriter()
        chunker_cite = DummyWriter()
    else:
        chunker_grant = ChunkWriter(opath_grant, schema=schema_grant)
        chunker_ipc = ChunkWriter(opath_ipc, schema=schema_ipc)
        chunker_cite = ChunkWriter(opath_cite, schema=schema_cite)

    if fname.endswith('.dat'):
        gen = 1
        parser = parse_grant_gen1
    elif fname.startswith('pgb'):
        gen = 2
        parser = lambda fp: parse_wrapper(fp, 'PATDOC', parse_grant_gen2)
    elif fname.startswith('ipgb'):
        gen = 3
        parser = lambda fp: parse_wrapper(fp, 'us-patent-grant', parse_grant_gen3)
    else:
        print(f'{ftag}: Unknown format')

    # parse it up
    try:
        print(f'{ftag}: Starting')

        i = 0
        for pat in parser(fpath):
            i += 1

            # store all info
            store_patent(pat, chunker_grant, chunker_ipc, chunker_cite)

            # output if needed
            if display > 0 and i % display == 0:
                spat = {k: pat.get(k, '') for k in schema_grant}
                print(
                    'pn = {patnum:10.10s}, pd = {pubdate:10.10s}, ti = {title:30.30s}, '
                    'on = {owner:30.30s}, ci = {city:15.15s}, st = {state:2s}, '
                    'ct = {country:2s}'.format(**spat)
                )

        # commit remaining
        chunker_grant.commit()
        chunker_ipc.commit()
        chunker_cite.commit()

        print(f'{ftag}: Parsed {i} patents')
    except Exception as e:
        print(f'{ftag}: EXCEPTION OCCURRED!')
        print_exc()

        chunker_grant.delete()
        chunker_ipc.delete()
        chunker_cite.delete()

# main entry point
def parse_many(files, output, threads=10, display=1_000, overwrite=False, dryrun=False):
    # needed for multiprocess
    global parse_file_opts

    # collect files
    if type(files) is str or isinstance(files, os.PathLike):
        file_list = (
            sorted(glob.glob(f'{files}/*.dat')) +
            sorted(glob.glob(f'{files}/pgb*.xml')) +
            sorted(glob.glob(f'{files}/ipgb*.xml'))
        )
    else:
        file_list = args.target

    # ensure output dir
    if not dryrun and not os.path.exists(output):
        print(f'Creating directory {output}')
        os.makedirs(output)

    # apply options
    def parse_file_opts(fpath):
        parse_file(fpath, output, display=display, overwrite=overwrite, dryrun=dryrun)

    # parse files
    with Pool(threads) as pool:
        pool.map(parse_file_opts, file_list, chunksize=1)



# name matching using locality-sensitive hashing (simhash)
# these are mostly idempotent

import os
import numpy as np
import pandas as pd
import networkx as nx
from math import ceil
from itertools import chain, repeat
from collections import defaultdict
from editdistance import eval as levenshtein

from ..tools.standardize import standardize_weak, standardize_strong
from ..tools.tables import read_csv
from ..tools.simhash import shingle, Cluster

# firm name sources - tag: (table, id_col, name_col)
colmap = {
    'apply': ('apply_apply', 'appnum', 'appname'),
    'grant': ('grant_grant', 'patnum', 'owner'),
    'assignor': ('assign_use', 'assignid', 'assignor'),
    'assignee': ('assign_use', 'assignid', 'assignee'),
    'compustat': ('compustat_compustat', 'compid', 'name'),
}
sources0 = ['apply', 'grant', 'assignee', 'assignor']

def get_columns(sources):
    sources = sources if sources is not None else sources0
    return {k: v for k, v in colmap.items() if k in sources}

# find all unique names
def generate_names(output, sources=None):
    print('generating names')

    columns = get_columns(sources)

    sdict = {}
    for tag, (table, id_col, name_col) in columns.items():
        ipath = os.path.join(output, f'{table}.csv')
        src = read_csv(ipath, usecols=[id_col, name_col])
        sdict[tag] = src.dropna().rename({name_col: 'name'}, axis=1)

    names = pd.concat([src['name'] for src in sdict.values()], axis=0)
    names = names.apply(standardize_weak).drop_duplicates()
    names = names[names.str.len()>0].reset_index(drop=True)
    names = names.rename('name').rename_axis('id').reset_index()
    names.to_csv(f'{output}/name.csv', index=False)

    for tag, (table, id_col, name_col) in columns.items():
        opath = os.path.join(output, f'{tag}_match.csv')
        src = pd.merge(sdict[tag], names, how='left', on='name')
        src['id'] = src['id'].astype('Int64')
        src[[id_col, 'id']].to_csv(opath, index=False)

    print(f'found {len(names)} names')

# k = 8, thresh = 4 works well
def filter_pairs(output, nshingle=2, k=8, thresh=4):
    print('filtering pairs')

    c = Cluster(k=k, thresh=thresh)
    name_dict = {}

    npath = os.path.join(output, 'name.csv')
    names = read_csv(npath, usecols=['id', 'name'])

    for i, id, name in names.itertuples():
        words = name.split()
        shings = list(shingle(name, nshingle))

        features = shings + words
        weights = list(np.linspace(1.0, 0.0, len(shings))) + list(np.linspace(1.0, 0.0, len(words)))

        c.add(features, weights=weights, label=id)
        name_dict[id] = name

        if i > 0 and i % 100_000 == 0:
            print(f'{i}: {len(c.unions)}')

    ppath = os.path.join(output, 'pair.csv')
    pairs = pd.DataFrame([
        (i1, i2, name_dict[i1], name_dict[i2]) for i1, i2 in c.unions
    ], columns=['id1', 'id2', 'name1', 'name2'])
    pairs.to_csv(ppath, index=False)

    print('Found %i pairs' % len(pairs))

# compute distances on owners in same cluster
def find_groups(output, thresh=0.85):
    print('finding matches')

    def dmetr(name1, name2):
        max_len = max(len(name1), len(name2))
        max_dist = int(ceil(max_len*(1.0-thresh)))
        ldist = levenshtein(name1, name2)
        return (1.0 - float(ldist)/max_len) if (ldist != -1 and max_len != 0) else 0.0

    close = []
    name_std = {}

    ppath = os.path.join(output, 'pair.csv')
    pairs = read_csv(ppath, usecols=['id1', 'id2', 'name1', 'name2'])

    for i, id1, id2, name1, name2 in pairs.itertuples():
        if id1 not in name_std:
            name_std[id1] = standardize_strong(name1)
        if id2 not in name_std:
            name_std[id2] = standardize_strong(name2)

        n1std = name_std[id1]
        n2std = name_std[id2]

        if dmetr(n1std, n2std) > thresh:
            close.append((id1, id2))

        if i > 0 and i % 100_000 == 0:
            print(f'{i}: {len(close)}')

    G = nx.Graph()
    G.add_edges_from(close)
    comps = sorted(nx.connected_components(G), key=len, reverse=True)

    mpath = os.path.join(output, 'match.csv')
    match = pd.DataFrame(chain(*[
        zip(repeat(fid), ids) for fid, ids in enumerate(comps)
    ]), columns=['firm_num', 'id'])
    match.to_csv(mpath, index=False)

    print(f'found {len(comps)} groups')

# must be less than 1000000 components
def merge_firms(output, sources=None, base=1000000):
    print('merging firms')

    columns = get_columns(sources)

    npath = os.path.join(output, 'name.csv')
    mpath = os.path.join(output, 'match.csv')
    fpath = os.path.join(output, 'firm.csv')

    names = read_csv(npath)
    match = read_csv(mpath)

    firms = pd.merge(names, match, how='left', on='id')
    firms['firm_num'] = firms['firm_num'].fillna(firms['id']+base).astype(np.int)
    firms[['firm_num', 'id']].to_csv(fpath, index=False)

    for tag, (table, id_col, name_col) in columns.items():
        ipath = os.path.join(output, f'{tag}_match.csv')
        opath = os.path.join(output, f'{tag}_firm.csv')
        src = read_csv(ipath)
        src = pd.merge(src, firms, on='id')
        src[[id_col, 'firm_num']].to_csv(opath, index=False)

# go through all steps
def cluster_firms(output, sources=None):
    generate_names(output, sources=sources)
    filter_pairs(output)
    find_groups(output)
    merge_firms(output, sources=sources)



#
# locally sensitive hashing code
#

from collections import defaultdict
import numpy as np
import xxhash

# add current path
import os
import sys
own_path, _ = os.path.split(__file__)
sys.path.insert(0, own_path)

# import cython
try:
    import pyximport
    pyximport.install()
    import simcore as simcore
except:
    pass

# k-shingles: pairs of adjacent k-length substrings (in order)
def shingle(s, k=2):
    k = min(len(s), k)
    for i in range(len(s)-k+1):
        yield s[i:i+k]

# split into words
def tokenize(s):
    return s.split()

def murmur(x):
    return np.uint64(xxhash.xxh64_intdigest(x))

# compute actual simhash
class Simhash:
    def __init__(self):
        self.dim = 64
        self.unums = list(map(np.uint64,range(self.dim)))
        self.masks = [self.unums[1] << n for n in self.unums]

    def simhash(self, features, weights=None):
        if weights is None:
            weights = [1.0]*len(features)
        hashish = [murmur(f) for f in features]
        v = [0.0]*self.dim
        for h, w in zip(hashish, weights):
            for i in range(self.dim):
                v[i] += w if h & self.masks[i] else -w
        ans = self.unums[0]
        for i in range(self.dim):
            if v[i] >= 0:
                ans |= self.masks[i]
        return ans

# compute actual simhash with C - only 64 width
class CSimhash():
    def __init__(self):
        self.simcore = simcore.simcore

    def simhash(self, features, weights=None):
        if weights is None:
            weights = [1.0]*len(features)
        hashish = [murmur(f) for f in features]
        ret = np.uint64(self.simcore(hashish, weights))
        return ret

class Cluster:
    # dim is the simhash width, k is the tolerance
    def __init__(self, dim=64, k=4, thresh=1):
        self.dim = dim
        self.k = k
        self.thresh = thresh

        self.unions = []
        self.hashmaps = [defaultdict(list) for _ in range(k)] # defaultdict(list)
        self.offsets = [np.uint64(dim//k*i) for i in range(k)]
        self.bin_masks = [np.uint64(2**(dim-offset)-1) if (i == len(self.offsets)-1) else np.uint64(2**(self.offsets[i+1]-offset)-1) for i, offset in enumerate(self.offsets)]

        self.csim = CSimhash()
        self.hasher = self.csim.simhash

    # add item to the cluster
    def add(self, features, label, weights=None):
        # get subkeys
        sign = self.hasher(features, weights)
        keyvec = self.get_keys(sign)

        # Unite labels with the same keys in the same band
        matches = defaultdict(int)
        for idx, key in enumerate(keyvec):
            others = self.hashmaps[idx][key]
            for l in others:
                matches[l] += 1
            others.append(label)
        for out, val in matches.items():
           if val > self.thresh:
               self.unions.append((label, out))

    # bin simhash into chunks
    def get_keys(self, simhash):
        return [simhash >> offset & mask for offset, mask in zip(self.offsets, self.bin_masks)]



DEF dim = 64

cdef unsigned long masks[dim]
for i in range(dim):
  masks[i] = 1 << i

cdef float v[dim]

def simcore(hashish, weights):
    cdef unsigned long long ans
    cdef int n = len(hashish)
    cdef unsigned long long h
    cdef float w
    cdef float q

    for j in range(dim):
        v[j] = 0.0

    for i in range(n):
        h = hashish[i]
        w = weights[i]
        for j in range(dim):
            if h & masks[j]:
              q = w
            else:
              q = -w
            v[j] += q

    ans = 0
    for j in range(dim):
        if v[j] >= 0:
            ans |= masks[j]

    return ans



import re
import os
import glob
from collections import defaultdict
from traceback import print_exc
from lxml.etree import iterparse
from multiprocessing import Pool

from ..tools.parse import *
from ..tools.tables import ChunkWriter, DummyWriter

# parse assignment
def parse_assign_gen3(elem, fname):
    pat = defaultdict(str)
    pat['gen'] = 3
    pat['file'] = fname

    # top-level section
    record = elem.find('assignment-record')
    assignor = elem.find('patent-assignors')[0]
    assignee = elem.find('patent-assignees')[0]
    patents = elem.find('patent-properties')

    # assign id = reel-no + frame-no
    pat['bulkid'] = get_text(record, 'reel-no') + get_text(record, 'frame-no').zfill(3)

    # conveyance
    pat['conveyance'] = get_text(record, 'conveyance-text')

    # names
    pat['assignor'] = get_text(assignor, 'name')
    pat['assignee'] = get_text(assignee, 'name')

    # dates
    pat['execdate'] = get_text(assignor, 'execution-date/date')
    pat['recdate'] = get_text(record, 'recorded-date/date')

    # location
    pat['assignee_country'] = get_text(assignee, 'country-name', default='united states')
    pat['assignee_state'] = get_text(assignee, 'state')

    # patent info
    pat['patnums'] = [prune_patnum(pn) for pn in gen3_assign(patents)]

    return pat

# parse file
def parse_file_gen3(fpath):
    _, fname = os.path.split(fpath)
    for event, elem in iterparse(fpath, tag='patent-assignment', events=['end'], recover=True):
        yield parse_assign_gen3(elem, fname)
        clear(elem)

# table schema
schema_assign = {
    'assignid': 'str', # Assign ID
    'bulkid': 'str', # Group ID
    'patnum': 'str', # Patent number
    'execdate': 'str', # Execution date
    'recdate': 'str', # Reception date
    'conveyance': 'str', # Conveyance description
    'assignor': 'str', # Assignor name
    'assignee': 'str', # Assignee name
    'assignee_state': 'str', # State code
    'assignee_country': 'str', # Assignee country
    'gen': 'int', # USPTO data format
    'file': 'str', # path to source file
}

# chunking express
def store_patent(pat, chunker_assign):
    # filter out individuals and non-transfers
    pat['assignor_type'] = org_type(pat['assignor'])
    pat['assignee_type'] = org_type(pat['assignee'])
    pat['convey_type'] = convey_type(pat['conveyance'])
    if pat['assignor_type'] == ORG_INDV or pat['assignee_type'] == ORG_INDV or pat['convey_type'] == CONV_OTHER:
        return

    # store assign
    for i, pn in enumerate(pat['patnums']):
        pat['patnum'] = pn
        pat['assignid'] = pat['bulkid'] + '_' + str(i)
        chunker_assign.insert(*(pat[k] for k in schema_assign))

# file level
def parse_file(fpath, output, display=0, overwrite=False, dryrun=False):
    fdir, fname = os.path.split(fpath)
    ftag, fext = os.path.splitext(fname)

    opath_assign = os.path.join(output, f'assign_{ftag}.csv')

    if not overwrite:
        if os.path.exists(opath_assign):
            print(f'{ftag}: Skipping')
            return

    if dryrun:
        chunker_assign = DummyWriter()
    else:
        chunker_assign = ChunkWriter(opath_assign, schema=schema_assign)

    # parse it up
    try:
        print(f'{ftag}: Starting')

        i = 0
        for pat in parse_file_gen3(fpath):
            i += 1

            store_patent(pat, chunker_assign)

            # output
            if display > 0 and i % display == 0:
                pat['npat'] = len(pat['patnums'])
                print(
                    '[{npat:4d}]: {assignor:40.40s} [{assignor_type:1d}] -> '
                    '{assignee:30.30s} [{assignee_type:1d}] ({recdate:8.8s}, '
                    '{assignee_country:20.20s})'.format(**pat)
                )

        print(f'{ftag}: Parsed {i} records')

        # clear out the rest
        chunker_assign.commit()
    except Exception as e:
        print(f'{ftag}: EXCEPTION OCCURRED!')
        print_exc()

        chunker_assign.delete()

# main entry point
def parse_many(files, output, threads=10, display=1_000, overwrite=False, dryrun=False):
    # needed for multiprocess
    global parse_file_opts

    # collect files
    if type(files) is str or isinstance(files, os.PathLike):
        file_list = sorted(glob.glob(f'{files}/*.xml'))
    else:
        file_list = files

    # ensure output dir
    if not dryrun and not os.path.exists(output):
        print(f'Creating directory {output}')
        os.makedirs(output)

    # apply options
    def parse_file_opts(fpath):
        parse_file(fpath, output, display=display, overwrite=overwrite, dryrun=dryrun)

    # parse files
    with Pool(threads) as pool:
        pool.map(parse_file_opts, file_list, chunksize=1)



import os
import glob
import numpy as np
import pandas as pd

# maint file layout
colspec = [(0, 13), (14, 22), (23, 24), (25, 33), (34, 42), (43, 51), (52, 56)]

# event code mapping
m4 = ['M1551', 'M170', 'M173', 'M183', 'M2551', 'M273', 'M283']
m8 = ['M1552', 'M171', 'M174', 'M184', 'M2552', 'M274', 'M284']
m12 = ['M1553', 'M172', 'M175', 'M185', 'M2553', 'M275', 'M285']
mmap = [(m, 4) for m in m4] + [(m, 8) for m in m8] + [(m, 12) for m in m12]
codes = pd.DataFrame(mmap, columns=['code', 'lag']).set_index('code')

def parse_file(fpath, output, overwrite=False, dryrun=False):
    fdir, fname = os.path.split(fpath)
    opath = os.path.join(output, 'maint_maint.csv')

    if not overwrite and os.path.exists(opath):
        print(f'{fname}: Skipping')
        return
    else:
        print(f'{fname}: Starting')

    # import to dataframe
    print('Reading table')
    datf = pd.read_fwf(
        fpath, colspecs=colspec, usecols=[0, 2, 6], names=['patnum', 'is_small', 'event_code']
    )
    datf['patnum'] = datf['patnum'].apply(lambda s: s.lstrip('0').lower())

    # clean up data
    print('Data cleanup')
    datf = datf.join(codes, on='event_code', how='left').dropna()
    datf = datf.drop('event_code', axis=1)
    datf['is_small'] = datf['is_small'] == 'Y'
    pat_groups = datf.groupby('patnum')
    dpat = pd.DataFrame({
        'last_maint': pat_groups['lag'].max().astype(int),
        'ever_large': ~pat_groups['is_small'].min().astype(bool)
    }).reset_index()

    # write to disk
    print('Writing table')
    if not dryrun:
        dpat.to_csv(opath, index=False)

def get_date(fpath):
    fdir, fname = os.path.split(fpath)
    fbase, _ = os.path.splitext(fname)
    _, date = fbase.split('_')
    return int(date)

# really this is only one file
def parse_many(files, output, overwrite=False, dryrun=False, threads=None):
    if os.path.isdir(files):
        # get latest file
        maint_files = glob.glob(f'{files}/MaintFeeEvents_*.txt')
        dates = [get_date(fp) for fp in maint_files]
        max_index = np.argmax(dates)
        file_one = maint_files[max_index]
    else:
        file_one = files

    # ensure output dir
    if not dryrun and not os.path.exists(output):
        print(f'Creating directory {output}')
        os.makedirs(output)

    parse_file(file_one, output, overwrite=overwrite, dryrun=dryrun)



# match citation data with aggregated firm data (to be run before firm merge)

import os
import numpy as np
import pandas as pd

from ..tools.tables import read_csv

# match and aggregates cites
def aggregate_chunk(cites, grants):
    print(len(cites))

    # match citations to firms with patnum
    cites = cites.rename(columns={'src': 'citer_pnum', 'dst': 'citee_pnum'})
    cites = cites.join(grants.add_prefix('citer_'), on='citer_pnum')
    cites = cites.join(grants.add_prefix('citee_'), on='citee_pnum')
    cites['self_cite'] = (cites['citer_firm_num'] == cites['citee_firm_num']).fillna(False)

    # patent level statistics
    stats = pd.DataFrame({
        'n_cited': cites.groupby('citer_pnum').size(),
        'n_citing': cites.groupby('citee_pnum').size(),
        'n_self_cited': cites.groupby('citer_pnum')['self_cite'].sum(),
        'n_self_citing': cites.groupby('citee_pnum')['self_cite'].sum(),
    }).rename_axis(index='patnum')
    stats = stats.fillna(0).astype(np.int)

    return stats

def aggregate_cites(output, chunksize=10000000):
    grant_path = os.path.join(output, 'grant_firm.csv')
    cite_path = os.path.join(output, 'grant_cite.csv')
    stats_path = os.path.join(output, 'cite_stats.csv')

    # load in grant data
    grants = read_csv(grant_path).set_index('patnum')
    request = read_csv(cite_path, chunksize=chunksize)

    # loop over citation chunks (otherwise requires >32GB of RAM)
    cite_stats = pd.concat([aggregate_chunk(df, grants) for df in request], axis=0)
    cite_stats = cite_stats.groupby('patnum').sum() # since patents can span multiple chunks
    cite_stats.to_csv(stats_path)



import re

#
# weak name standardization
#

# regular expression substitutions
paren = r'\'s|\(.*\)|\.'
punct = r'[^\w\s]'
space = r'[ ]{2,}'

paren_re = re.compile(paren)
punct_re = re.compile(punct)
space_re = re.compile(space)

# standardize firm name
def standardize_weak(name):
    name = name.lower()
    name = paren_re.sub(' ', name)
    name = punct_re.sub(' ', name)
    name = space_re.sub(' ', name)
    return name.strip()

#
# strong name standardization
#

# whitespace
white0 = r' +'
white0_re = re.compile(white0)

# postscripts
post0 = r'(\ba corp.|;|,).*$'
post0_re = re.compile(post0)

# acronyms
acronym1 = r'\b(\w) (\w) (\w)\b'
acronym1_re = re.compile(acronym1)
acronym2 = r'\b(\w) (\w)\b'
acronym2_re = re.compile(acronym2)
acronym3 = r'\b(\w)-(\w)-(\w)\b'
acronym3_re = re.compile(acronym3)
acronym4 = r'\b(\w)-(\w)\b'
acronym4_re = re.compile(acronym4)
acronym5 = r'\b(\w\w)&(\w)\b'
acronym5_re = re.compile(acronym5)
acronym6 = r'\b(\w)&(\w)\b'
acronym6_re = re.compile(acronym6)
acronym7 = r'\b(\w) & (\w)\b'
acronym7_re = re.compile(acronym7)

# punctuation
punct0 = r'\'s|\(.*\)|\.'
punct1 = r'[^\w\s]'
punct0_re = re.compile(punct0)
punct1_re = re.compile(punct1)

# generic terms
states = ['del', 'de', 'ny', 'va', 'ca', 'pa', 'oh', 'nc', 'wi', 'ma']
compustat = ['plc', 'cl', 'redh', 'adr', 'fd', 'lp', 'cp', 'tr', 'sp', 'cos', 'gp', 'old', 'new']
generics = ['the', 'a', 'of', 'and', 'an']
en_corps = ['corporation', 'incorporated', 'company', 'limited', 'inc', 'llc', 'ltd', 'corp']
eu_corps = ['aktiengesellschaft', 'aktiebolag', 'ag', 'nv', 'bv', 'gmbh', 'co', 'bv', 'sa', 'ab', 'se']
jp_corps = ['kabushiki', 'kaisha', 'seisakusho']
variants = ['trust', 'group', 'grp', 'hldgs', 'holdings', 'comm', 'inds', 'hldg', 'tech', 'international', 'comp']
dropout = states + compustat + generics + en_corps + eu_corps + jp_corps + variants
gener0_re = re.compile('|'.join([rf'\b{el}\b' for el in dropout]))

# standardize a firm name
def standardize_strong(name):
    name = name.lower()
    name = post0_re.sub('', name)
    name = acronym1_re.sub(r'\1\2\3', name)
    name = acronym2_re.sub(r'\1\2', name)
    name = acronym3_re.sub(r'\1\2\3', name)
    name = acronym4_re.sub(r'\1\2', name)
    name = acronym5_re.sub(r'\1\2', name)
    name = acronym6_re.sub(r'\1\2', name)
    name = acronym7_re.sub(r'\1\2', name)
    name = punct0_re.sub('', name)
    name = punct1_re.sub(' ', name)
    name = gener0_re.sub('', name)
    name = white0_re.sub(' ', name)
    return name.strip()



import os
import glob
import argparse
from itertools import chain

tables = {
    'grant': ['grant', 'ipc', 'cite'],
    'apply': ['apply', 'ipc'],
    'assign': ['assign'],
    'maint': ['maint'],
    'tmapply': ['tmapply'],
    'compustat': ['compustat'],
}

def concat_files(input, output, case, table, dryrun=False):
    system = print if dryrun else os.system

    fout = f'{output}/{case}_{table}.csv'
    files = sorted(glob.glob(f'{input}/{table}_*.csv'))

    if len(files) == 0:
        print(f'Table "{case}/{table}" not found')
        return

    first = files[0]
    flist = ' '.join(files)

    system(f'head -n 1 {first} > {fout}')
    system(f'tail -q -n +2 {flist} >> {fout}')

def concat_tables(input, output, ftype, dryrun=False):
    if not dryrun and not os.path.exists(output):
        print(f'Creating directory {output}')
        os.makedirs(output)

    for tab in tables[ftype]:
        print(f'Concat: {ftype}/{tab}')
        concat_files(input, output, ftype, tab, dryrun=dryrun)



# tools for patent data

import os
import pandas as pd

# csv types
dtypes = {
    'patnum': 'str',
    'appnum': 'str',
    'appdate': 'str',
    'pubdate': 'str',
    'execdate': 'str',
    'first_trans': 'str',
    'naics': 'Int64',
    'sic': 'Int64',
    'appname': 'str',
    'owner': 'str',
    'assignor': 'str',
    'assignee': 'str',
    'firm_num': 'Int64',
    'assignid': 'str',
    'bulkid': 'str',
    'id': 'Int64',
    'id1': 'Int64',
    'id2': 'Int64',
    'name': 'str',
    'name1': 'str',
    'name2': 'str',
    'src': 'str',
    'dst': 'str',
}

# read csv with proper types
def read_csv(fname, **kwargs):
    dt = {**dtypes, **kwargs.pop('dtype', {})}
    return pd.read_csv(fname, dtype=dt, **kwargs)

def astype(data, dtype):
    if dtype == 'str':
        return pd.Series(data, dtype='str')
    elif dtype == 'int':
        return pd.to_numeric(pd.Series(data), errors='coerce').astype('Int64')
    else:
        raise Exception(f'Unsupported type: {dtype}')

# insert in chunks
class ChunkWriter:
    def __init__(self, path, schema, chunk_size=1000, output=False):
        self.path = path
        self.schema = schema
        self.chunk_size = chunk_size
        self.output = output
        self.items = []
        self.i = 0
        self.j = 0

        self.file = open(self.path, 'w+')
        header = ','.join(schema)
        self.file.write(f'{header}\n')

    def __del__(self):
        self.file.close()

    def insert(self, *args):
        self.items.append(args)
        if len(self.items) >= self.chunk_size:
            self.commit()
            return True
        else:
            return False

    def insertmany(self, args):
        self.items += args
        if len(self.items) >= self.chunk_size:
            self.commit()
            return True
        else:
            return False

    def commit(self):
        self.i += 1
        self.j += len(self.items)

        if len(self.items) == 0:
            return

        if self.output:
            print(f'Committing chunk {self.i} to {self.table} ({len(self.items)})')

        data = [x for x in zip(*self.items)]
        frame = pd.DataFrame({k: astype(d, v) for (k, v), d in zip(self.schema.items(), data)})
        frame.to_csv(self.file, index=False, header=False)

        self.items.clear()

    def delete(self):
        self.file.close()
        os.remove(self.path)

# pretend to insert in chunks
class DummyWriter:
    def __init__(self, *args, chunk_size=1000, output=False, **kwargs):
        self.chunk_size = chunk_size
        self.output = output
        self.last = None
        self.i = 0

    def insert(self, *args):
        self.last = args
        self.i += 1
        if self.i >= self.chunk_size:
            self.commit()
            return True
        else:
            return False

    def insertmany(self, args):
        if len(args) > 0:
            self.last = args[-1]
        self.i += len(args)
        if self.i >= self.chunk_size:
            self.commit()
            return True
        else:
            return False

    def commit(self):
        if self.output:
            print(self.last)
        self.i = 0

    def delete(self):
        pass



import os
import pandas as pd

from ..tools.standardize import standardize_strong
from ..tools.tables import read_csv

# detect same entity transfers
def same_entity(assignor, assignee):
    assignor_toks = standardize_strong(assignor)
    assignee_toks = standardize_strong(assignee)

    word_match = 0
    for tok in assignor_toks:
        if tok in assignee_toks:
            word_match += 1

    word_match /= max(1.0, 0.5*(len(assignor_toks)+len(assignee_toks)))
    match = word_match > 0.5
    return match

# map to two-char state codes
state_map = {
    '': '',
    'alabama': 'al',
    'alaska': 'ak',
    'arizona': 'az',
    'arkansas': 'ar',
    'california': 'ca',
    'colorado': 'co',
    'connecticut': 'ct',
    'delaware': 'de',
    'district of columbia': 'dc',
    'florida': 'fl',
    'georgia': 'ga',
    'hawaii': 'hi',
    'idaho': 'id',
    'illinois': 'il',
    'indiana': 'in',
    'iowa': 'ia',
    'kansas': 'ks',
    'kentucky': 'ky',
    'louisiana': 'la',
    'maine': 'me',
    'maryland': 'md',
    'massachusetts': 'ma',
    'michigan': 'mi',
    'minnesota': 'mn',
    'mississippi': 'ms',
    'missouri': 'mo',
    'montana': 'mt',
    'nebraska': 'ne',
    'nevada': 'nv',
    'new hampshire': 'nh',
    'new jersey': 'nj',
    'new mexico': 'nm',
    'new york': 'ny',
    'north carolina': 'nc',
    'north dakota': 'nd',
    'ohio': 'oh',
    'oklahoma': 'ok',
    'oregon': 'or',
    'pennsylvania': 'pa',
    'rhode island': 'ri',
    'south carolina': 'sc',
    'south dakota': 'sd',
    'tennessee': 'tn',
    'texas': 'tx',
    'utah': 'ut',
    'vermont': 'vt',
    'virginia': 'va',
    'washington': 'wa',
    'west virginia': 'wv',
    'wisconsin': 'wi',
    'wyoming': 'wy'
}

# map to two-char country codes
country_map = {
    '': '',
    'united states': 'us',
    'singapore': 'sg',
    'japan': 'jp',
    'cayman islands': 'ky',
    'korea, republic of': 'kr',
    'netherlands': 'nl',
    'germany': 'de',
    'ireland': 'ie',
    'switzerland': 'ch',
    'china': 'cn',
    'france': 'fr',
    'canada': 'ca',
    'united kingdom': 'uk',
    'finland': 'fi',
    'luxembourg': 'lu',
    'taiwan': 'tw',
    'puerto rico': 'us',
    'hong kong': 'hk',
    'great britain': 'uk',
    'sweden': 'se',
    'stateless': '',
    'england': 'uk',
    'barbados': 'bb',
    'bermuda': 'bm',
    'italy': 'it',
    'denmark': 'dk',
    'israel': 'il',
    'australia': 'au',
    'ontario': 'ca',
    'belgium': 'be',
    'not provided': '',
    'virgin islands, british': 'vg',
    'cyprus': 'cy',
    'austria': 'at',
    'norway': 'no',
    'russian federation': 'ru',
    'india': 'in',
    'spain': 'es',
    'korea, democratic people\'s republic of': 'kp',
    'new zealand': 'nz',
    'guernsey': 'gg',
    'samoa': 'ws',
    'saudi arabia': 'sa',
    'mexico': 'mx',
    'malaysia': 'my',
    'brazil': 'br',
    'south africa': 'za',
    'scotland': 'uk',
    'hungary': 'hu',
    'netherlands antilles': 'an',
    'british columbia': 'ca',
    'liechtenstein': 'li',
    'seychelles': 'sc',
    'portugal': 'pt',
    'quebec': 'ca',
    'united arab emirates': 'ae',
    'german democratic republic': 'de',
    'malta': 'mt',
    'isle of man': 'im',
    'panama': 'pa',
    'bahamas': 'bs',
    'iceland': 'is',
    'georgia': 'ge',
    'channel islands': 'uk',
    'alberta': 'ca',
    'thailand': 'th',
    'czech republic': 'cz',
    'mauritius': 'mu',
    'philippines': 'ph',
    'namibia': 'na',
    'chile': 'cl',
    'saint kitts and nevis': 'kn',
    'jersey': 'je',
    'argentina': 'ar',
    'turks and caicos islands': 'tc',
    'gibraltar': 'gi',
    'manitoba': 'ca',
    'belize': 'bz',
    'wales': 'uk',
    'turkey': 'tr',
    'greece': 'gr',
    'slovakia': 'sk',
    'virgin islands, u.s.': 'us',
    'cook islands': 'ck',
    'poland': 'pl',
    'macao': 'mo',
    'nova scotia': 'ca',
    'anguilla': 'ai',
    'colombia': 'co',
    'iran, islamic republic of': 'ir',
    'monaco': 'mc',
    'marshall islands': 'mh',
    'croatia': 'hr',
    'sri lanka': 'lk',
    'venezuela': 've',
    'jordan': 'jo',
    'brunei darussalam': 'bn',
    'saskatchewan': 'ca',
    'uruguay': 'uy',
    'estonia': 'ee',
    'ukraine': 'ua',
    'northern ireland': 'uk',
    'cuba': 'cu',
    'liberia': 'lr',
    'kuwait': 'kw',
    'qatar': 'qa',
    'san marino': 'sm',
    'romania': 'ro',
    'cocos (keeling) islands': 'cc',
    'slovenia': 'si',
    'antigua and barbuda': 'ag',
    'swaziland': 'sz',
    'vanuatu': 'vu',
    'papua new guinea': 'pg',
    'latvia': 'lv',
    'union of soviet socialist republics': 'ru',
    'curacao': 'cw',
    'costa rica': 'cr',
    'viet nam': 'vn',
    'england and wales': 'uk',
    'norfolk island': 'au',
    'macau': 'mo',
    'jamaica': 'jm',
    'bulgaria': 'bg',
    'european union': 'eu',
    'trinidad and tobago': 'tt',
    'newfoundland and labrador': 'ca',
    'armenia': 'am',
    'czechoslovakia': 'cz',
    'guam': 'gu',
    'andorra': 'ad',
    'algeria': 'dz',
    'indonesia': 'id',
    'belarus': 'by',
    'azerbaijan': 'az',
    'dominica': 'dm',
    'peru': 'pe',
    'dominican republic': 'do',
    'aruba': 'aw',
    'british indian ocean territory': 'io',
    'oman': 'om',
    'guadeloupe': 'fr',
    'ecuador': 'ec',
    'lithuania': 'lt',
    'lebanon': 'lb',
    'prince edward island': 'ca',
    'kazakstan': 'kz',
    'botswana': 'bw',
    'christmas island': 'cx',
    'pakistan': 'pk',
    'american samoa': 'as',
    'saint lucia': 'lc',
    'saint vincent and the grenadines': 'vc',
    'tunisia': 'tn',
    'grenada': 'gd',
    'serbia': 'rs',
    'benin': 'bj',
    'mauritania': 'mr',
    'morocco': 'ma',
    'saint helena': 'sh',
    'bangladesh': 'bd',
    'french polynesia': 'pf',
    'chad': 'td',
    'yemen': 'ye',
    'syrian arab republic': 'sy',
    'nothern mariana islands': 'mp',
    'uzbekistan': 'uz',
    'rwanda': 'rw',
    'guatemala': 'gt',
    'new brunswick': 'ca',
    'burundi': 'bi'
}

def prune_assign(output):
    spath = os.path.join(output, 'assign_assign.csv')
    assn = read_csv(spath)

    # eliminate duplicate records - assume chronological order
    assn = assn.drop_duplicates('assignid', keep='last')
    assn = assn.dropna(subset=['patnum', 'execdate'], axis=0)

    # eliminate assignments within entities
    assn['assignee_state'] = assn['assignee_state'].map(state_map)
    assn['assignee_country'] = assn['assignee_country'].map(country_map)
    assn['same'] = assn[['assignor', 'assignee']].apply(
        lambda x: same_entity(*x), raw=True, axis=1
    )
    good = assn[~assn['same']].drop('same', axis=1)

    gpath = os.path.join(output, 'assign_use.csv')
    good.to_csv(gpath, index=False)

    # aggregated assignment stats
    pat_group = good.groupby('patnum')
    assign_stats = pd.DataFrame({
        'first_trans': pat_group['execdate'].min(),
        'n_trans': pat_group.size()
    }).reset_index()

    apath = os.path.join(output, 'assign_stats.csv')
    assign_stats.to_csv(apath, index=False)



import os
import pandas as pd

colmap = {
    'row': 'row',
    'gvkey': 'gvkey',
    'fyear': 'year',
    'conm': 'name',
    'AT': 'assets',
    'CAPX': 'capx',
    'CH': 'cash',
    'COGS': 'cogs',
    'CSHO': 'shares',
    'DP': 'deprec',
    'EBITDA': 'income',
    'EMP': 'employ',
    'INTAN': 'intan',
    'LT': 'debt',
    'REVT': 'revenue',
    'SALE': 'sales',
    'XRD': 'rnd',
    'XSGA': 'fcost',
    'prcc_f': 'price',
    'NAICS': 'naics',
    'SIC': 'sic'
}

dtype = {
    'gvkey': 'Int64',
    'year': 'Int64',
    'naics': 'Int64',
    'sic': 'Int64'
}

def parse_file(fpath, output, display=0, overwrite=False, dryrun=False):
    fdir, fname = os.path.split(fpath)
    opath = os.path.join(output, 'compustat_compustat.csv')

    if not overwrite and os.path.exists(opath):
        print(f'{fname}: Skipping')
        return
    else:
        print(f'{fname}: Starting')

    # read frame into memory
    datf = pd.read_csv(fpath, index_col=0, usecols=colmap, dtype=dtype)
    datf = datf.rename(columns=colmap)

    # clean up data
    datf['mktval'] = datf['shares']*datf['price']
    datf = datf.drop(['shares', 'price'], axis=1)
    datf = datf.dropna(subset=['gvkey', 'year', 'name'])
    datf['name'] = datf['name'].str.lower()
    datf['gvkey'] = datf['gvkey'].astype('Int64')
    datf['year'] = datf['year'].astype('Int64')
    datf['naics'] = datf['naics'].fillna(0).astype(int).map(lambda x: f'{x:<06d}')
    datf['sic'] = datf['naics'].fillna(0).astype(int).map(lambda x: f'{x:<06d}')
    datf = datf.reset_index(drop=True).rename_axis('compid').reset_index()

    # write to disk
    if not dryrun:
        datf.to_csv(opath, index=False, float_format='%.3f')

# really this is only one file
def parse_many(files, output, overwrite=False, dryrun=False, threads=None):
    if os.path.isdir(files):
        file_one = os.path.join(files, 'compustat.csv')
    else:
        file_one = files

    # ensure output dir
    if not dryrun and not os.path.exists(output):
        print(f'Creating directory {output}')
        os.makedirs(output)

    parse_file(file_one, output, overwrite=overwrite, dryrun=dryrun)



import os
import time

def fetch_file(zurl, output, overwrite=False, dryrun=False, unzip=False):
    system = print if dryrun else os.system
    zflags = '' if overwrite else '-n'

    if not dryrun and not os.path.exists(output):
        print(f'Creating directory {output}')
        os.makedirs(output)

    _, zname = os.path.split(zurl)
    zpath = os.path.join(output, zname)
    fetch = overwrite or not os.path.isfile(zpath)

    if fetch:
        print(f'Fetching {zname}')
        system(f'curl -o {zpath} {zurl}')

    if fetch or unzip:
        print(f'Unzipping {zname}')
        system(f'unzip {zflags} {zpath} -d {output}')

    return fetch

def fetch_many(files, output, delay=10, dryrun=False, **kwargs):
    for zurl in files:
        fetch = fetch_file(zurl, output, dryrun=dryrun, **kwargs)
        if fetch and not dryrun:
            time.sleep(delay)



from .tools.fetch import fetch_many
from .tools.concat import concat_tables

from .parse.apply import parse_many as parse_apply
from .parse.grant import parse_many as parse_grant
from .parse.assign import parse_many as parse_assign
from .parse.maint import parse_many as parse_maint
from .parse.tmapply import parse_many as parse_tmapply
from .parse.compustat import parse_many as parse_compustat

from .firms.cluster import cluster_firms
from .firms.assign import prune_assign
from .firms.cites import aggregate_cites
from .firms.merge import merge_firms

from . import tools
from . import parse
from . import firms
